class AgentosException(Exception):
    pass


class PythonComponentSystemException(Exception):
    pass


class BadGitStateException(Exception):
    pass


class RegistryException(Exception):
    pass
